#include "atari.h"
#include "config.h"
#include "monitor.h"
#include "colours.h"
#include "screen.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "atari_gp32.h"
#include "colour_hsv.h"
#include "gp2xminlib.h"

//#define COUNT_FPS
#ifdef COUNT_FPS
#define DEBUG_TEXT
#endif
//#define DEBUG_TEXT
//#define PALETTE

#ifdef NATY
	extern SDL_Surface * surface;
#endif

// Used externally as an easy way of getting text on the screen!
char * debugTextBuffer = 0;
char * debugText =0;

// Brightness (used externally in state save)
unsigned char defaultBrightness = 2;
unsigned char brightness = 2;

// Screen buffers -> theoretically we double buffer.
// In practice it is not bad without so it is off ATM
// I have performance concerns with GpSwapBuffer
// Need to investigate further
// Update: Just tried -> performance reduced by approx 30-40%
// WTF? Guess it must be waiting for sync?
int pal[256];

// Interrupt driven frame count
// Every second we get called back and store the value here
#ifdef COUNT_FPS
static char countFPSResult[256];
static int drawFPS = 0;
static int realFPS = 0;
static int realFPSCount = 0;
static int drawFPSCount = 0;
void fpsCallback() { drawFPS = drawFPSCount; drawFPSCount = 0; realFPS = realFPSCount; realFPSCount = 0; }
void countFPS(int fullRender);
#endif

// Notes on DIRTYRECT
// Tried it and found it to be slower than without.
// Seemed to be that when it was enabled the antic rendering took 10ms longer. Since the copy to screen takes 5ms that isn't a gain.

#ifndef PI
#define PI 3.142
#endif

///////////////////////////////////////////////////////////////
// Interface, called externally
void * Atari_DisplayActiveBuffer()
{
#ifdef PALETTE
	return gp2x_screen8;
#else
	return gp2x_screen15;
#endif
}

void Atari_DisplayInitPhase2()
{
#ifdef PALETTE
	gp2x_video_init(8,336,240,3);
#else
	gp2x_video_init(16,336,240,3);
#endif
	Atari_DisplayBrightness(brightness);

	gp2x_fill_rect(0,0,336,240,0);
	gp2x_video_flip();
}

void Atari_DisplayInitialise()
{
#ifdef DEBUG_TEXT
	debugTextBuffer = (char *) malloc(256);
#endif

	//// Set palette (colortable is setup by emulator palette code)
	//Atari_DisplayBrightness(brightness);

#ifdef COUNT_FPS
	// FPS count
	GpTimerOptSet(gpNextTimer, 1, 0, fpsCallback);
	GpTimerSet(gpNextTimer++);
#endif
}

void Atari_DisplayBrightness(int level)
{
#ifdef NATY
	SDL_Color entry[256];
#endif
	int i;

	brightness = level;

	for( i = 0 ; i < 256 ; i++ )
	{
		int R = (colortable[i]&0xff0000) >> 16;
		int G = (colortable[i]&0xff00) >> 8;
		int B = (colortable[i]&0xff) >> 0;

		if (brightness!=0)
		{
			double Rd=R,Gd=G,Bd=B;

			double Vadj = 1.0 + brightness/10.0;
			double Sadj = 1.0 + brightness/35.0;
			
			double H,S,V;
			rgb2hsv(Rd,Gd,Bd,&H,&S,&V);
			//fprintf(stderr, "BE:R:%03f:G:%03f:B:%03f H:%03f:S:%03f:V:%03f\n", Rd,Gd,Gd,H,S,V);
			V*=Vadj;
			S*=Sadj;
			if (V>1.0) V=1.0;
			if (S>1.0) S=1.0;
			hsv2rgb(H,S,V,&Rd,&Gd,&Bd);
                        //fprintf(stderr, "AF:R:%03f:G:%03f:B:%03f H:%03f:S:%03f:V:%03f\n", Rd,Gd,Gd,H,S,V);

			// Scale by sin first, to boost mid range
			//Rd = (1.0 + PI*(double)Rd/(double)256)/2.0;
			//Gd = (1.0 + PI*(double)Gd/(double)256)/2.0;
			//Bd = (1.0 + PI*(double)Bd/(double)256)/2.0;

			R = Rd<256 ? Rd : 255;
			G = Gd<256 ? Gd : 255;
			B = Bd<256 ? Bd : 255;
		}

#ifdef PALETTE
#ifdef NATY		
		entry[i].unused = 0;
		entry[i].r = R;
		entry[i].g = G;
		entry[i].b = B;
#endif
#endif
		pal[i] = (R<<16)|(G<<8)|B;
	}

	for (i=0; i<256; ++i)
	{
#ifdef PALETTE
		gp2x_video_color8(i,(pal[i]&0xFF0000)>>16, (pal[i]&0xFF00)>>8, (pal[i]&0xFF)); 
#endif
		pal[i] = gp2x_video_color15((pal[i]&0xFF0000)>>16, (pal[i]&0xFF00)>>8, (pal[i]&0xFF), 0);
	}

#ifdef PALETTE
#ifdef NATY
	SDL_SetPalette(surface, SDL_PHYSPAL, entry, 0, 256);
#endif
	gp2x_video_setpalette();
#endif
}

void Atari_DisplayScreen()
{	
	// Passed a null pointer if we are not expected to draw
	// Called every frame regardless...
#ifdef DEBUG_TEXT
	int oldbackbuf = backbuf;
#endif

	if (timeDrawNextFrame)
	{
		unsigned char * screen = (unsigned char *)atari_screen;
	//	if (SDL_MUSTLOCK(gp2Draw))
	//		SDL_LockSurface(gp2Draw);
		
		int width = 336-menuWidth;
		int offset = menuWidth + ((ATARI_WIDTH-336)/2);
		int height = 240-menuHeight;

#ifdef PALETTE
		char * dest = ((char *) gp2x_screen8) + menuWidth;
#else
		unsigned int * dest = (unsigned int *)(((short *) gp2x_screen15) + menuWidth);
#endif
		unsigned char * src = (unsigned char *) (screen + offset);

		int y=0;
		for (y=0; y<height; ++y)
		{
#ifdef PALETTE
			memcpy(dest,src,width);
			src+=ATARI_WIDTH;
			dest+=336;
#else
			int x=0;
			for (x=0; x<width; x+=2)
			{
				dest[x>>1] = pal[src[x+1]]<<16|pal[src[x]]; 
			}
			src += ATARI_WIDTH;
			dest += 336/2;
#endif
		}

		//if (SDL_MUSTLOCK(gp2Draw))
		//	SDL_UnlockSurface(gp2Draw);
		//SDL_UpdateRect(gp2Draw,0,0,0,0);
		gp2x_video_flip();
	}

//	debugText = debugTextBuffer;
//	sprintf(debugTextBuffer, "RTC:%8d %8d\n", rtcTick, GpTickCountGet());

#ifdef DEBUG_TEXT
	if (debugText)
	{
		GpRectFill(NULL, &gpDraw[oldbackbuf],5,220,305,10, 0x00);
		GpTextOut(0, &gpDraw[oldbackbuf], 9, 220, debugText, 0xff);
	}
#endif

#ifdef COUNT_FPS
	countFPS(screen!=0);
	GpRectFill(NULL, &gpDraw[oldbackbuf],5,230,305,10, 0x00);
	GpTextOut(0, &gpDraw[oldbackbuf], 9, 230, countFPSResult, 0xff);
#endif
}

///////////////////////////////////////////////////////////////
// Internal
#ifdef COUNT_FPS
void countFPS(int fullRender)
{
	static int first = 1;
	static int lastTicks = 0;
	static int totalTicks = 0;
	static int totalFrames = 0;
	static int skipCount = 0;

	static int fullRenderTicks[10];
	static int fullRenderPos = 0;

	if (first)
	{
		lastTicks = GpTickCountGet();
		first = 0;
	}
	else
	{
		int now = GpTickCountGet();
		int taken = now - lastTicks;
		int avgFrames = ((++totalFrames)*1000)/totalTicks;

		lastTicks = now;
		totalTicks += taken;

		if (fullRender) // Did we render this frame?
		{
			fullRenderPos++;
			if (fullRenderPos == 10)
				fullRenderPos = 0;

			fullRenderTicks[fullRenderPos] = taken;

			drawFPSCount++;
		}
		else
			skipCount++;
		realFPSCount++;

		int fullRenderTickCount=0;
		int k;
		for (k=0;k<10;++k)
			fullRenderTickCount += fullRenderTicks[k];

		sprintf(countFPSResult, "dps:%2d eps:%2d skp:%d rms:%2d rfp:%2d", drawFPS, realFPS, skipCount, fullRenderTickCount/10, 10000/fullRenderTickCount);

	}
}
#endif

int mapPal(int input)
{
#ifdef PALETTE
	return input;
#else
	return pal[input];
#endif
}

