#ifdef NATY
/* 
 * SDL version of minlib...
*/

#include "gp2xminlib.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <SDL/SDL.h>


enum  { GP2X_UP=0x1,       GP2X_LEFT=0x4,       GP2X_DOWN=0x10,  GP2X_RIGHT=0x40,
        GP2X_START=1<<8,   GP2X_SELECT=1<<9,    GP2X_L=1<<10,    GP2X_R=1<<11,
        GP2X_A=1<<12,      GP2X_B=1<<13,        GP2X_X=1<<14,    GP2X_Y=1<<15,
        GP2X_VOL_DOWN=1<<22, GP2X_VOL_UP=1<<23, GP2X_PUSH=1<<27, };

  unsigned int gp2x_bpp=0;
  unsigned long   gp2x_dev[8]={-1,-1,-1,-1,-1,-1,-1,-1}, gp2x_physvram[4], *gp2x_memregl;
  volatile unsigned long   gp2x_sound_pause=0, gp2x_sound=0, gp2x_sound_bits=0;
  unsigned short gp2x_sound_buffer16[8192];
  unsigned char gp2x_sound_buffer8[8192];
  volatile unsigned short  gp2x_palette[512];
  SDL_Surface * surface = 0;
  unsigned char * gp2x_screen8;
  unsigned short * gp2x_screen15;

void gp2x_video_flip(void)
{
	SDL_Flip(surface);
}
 
void gp2x_video_setpalette(void)
{
}

void gp2x_fill_rect(unsigned short x, unsigned short y, unsigned short width, unsigned short height, unsigned short col)
{
	if (x+width > 320)
		width = 320-x;
	if (y+height > 240)
		height = 240-y;

	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	rect.w = width;
	rect.h = height;
	SDL_FillRect(surface, &rect, col);
}

void gp2x_video_plot(unsigned short x, unsigned short y, unsigned short col)
{
	//determine position
	char* pPosition = ( char* ) surface->pixels ;

	//offset by y
	pPosition += ( surface->pitch * y ) ;

	//offset by x
	pPosition += ( surface->format->BytesPerPixel * x ) ;

	//copy pixel data
	*pPosition = col;
}

unsigned long gp2x_joystick_read(void)
{
	// unused
  return 0;
}


void gp2x_sound_play()
{
    audio_buf_info abi;
    while (ioctl(gp2x_dev[3], SNDCTL_DSP_GETOSPACE, &abi) || abi.fragments<1)
      usleep(1000);		     

    struct timespec ts;
    ts.tv_sec=0;
    ts.tv_nsec=10000000;
		    
    while (1)//abi.fragments--)
    {
      if (gp2x_sound_bits==8)
      {
       	while (!gp2x_sound_frame(0, (void *)gp2x_sound_buffer8, abi.fragsize/2)) nanosleep(&ts);
	int i=0;
	int end = abi.fragsize/2;
	for (i=0; i!=end; ++i)
	{
		unsigned short out = ((unsigned short)gp2x_sound_buffer8[i]);
		//fprintf(stderr,"O:%d\n",out);
		gp2x_sound_buffer16[i] = out<<7;

		int result = gp2x_sound_buffer16[i];

    static int min=1000000;
    static int min2=1000000;
    static int max=-1000000;
    static int max2=-1000000;
    static int ct=0;

    if (result<min) min=result;
    if (result>max) max=result;
    if (out<min2) min2=out;
    if (out>max2) max2=out;
    ++ct;
    if (ct>10000)
    {
	    ct = 0;
	    fprintf(stderr,"OUTmin:%d,%d max:%d,%d\n", min,min2,max,max2);
	    min = 1000000;
	    max = -min;
	    min2= 1000000;
	    max2 = -min2;
    }
	}
      }
      else
      {
        while (!gp2x_sound_frame(0, (void *)gp2x_sound_buffer16, abi.fragsize)) nanosleep(&ts);	      
      }
      write(gp2x_dev[3], (void *)gp2x_sound_buffer16, abi.fragsize);
    }
}


void gp2x_deinit(void)
{
  close(gp2x_dev[0]);
  close(gp2x_dev[1]);
  close(gp2x_dev[3]);
  close(gp2x_dev[4]);

  chdir("/usr/gp2x");
  execl("gp2xmenu",NULL);
  exit(0);
}

void gp2x_reg_init()
{ 
  gp2x_dev[0] = gp2x_dev[1] = gp2x_dev[3] = -1;	
}

void gp2x_video_init(int bpp)
{
  gp2x_bpp = bpp;

  SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);

  if (bpp == 8)
  {
  	surface = SDL_SetVideoMode(320,240,8,SDL_HWSURFACE|SDL_HWPALETTE);
	gp2x_screen8 = surface->pixels; 
  }
  else
  {
    	surface = SDL_SetVideoMode(320,240,16,SDL_HWSURFACE);
	gp2x_screen15 = surface->pixels;
  }
}

void gp2x_sound_init(int rate, int bits, int stereo)
{
  gp2x_dev[3] = open("/dev/dsp", O_WRONLY);
  gp2x_dev[4] = open("/dev/mixer", O_RDONLY);
  gp2x_sound_bits = bits;

//  int bits16=16; //nothing else allowed!
 int bits16= 0x80;
  ioctl(gp2x_dev[3], SNDCTL_DSP_SPEED,  &rate);
  ioctl(gp2x_dev[3], SNDCTL_DSP_SETFMT, &bits16);
  ioctl(gp2x_dev[3], SNDCTL_DSP_STEREO, &stereo);

  int frags = 0x1f0008;
  ioctl(gp2x_dev[3], SNDCTL_DSP_SETFRAGMENT, &frags);
}
#endif

