#include <gpstdio.h>
#include <gpstdlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gpdef.h>
#include <gpstdlib.h>
#include <gpgraphic.h>
#include <gpmm.h>
#include <gpstdio.h>
#include <gpfont.h>

#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/unistd.h>

#include <errno.h>
#undef errno
extern int errno;

extern GPDRAWSURFACE testSurface;
char testPrintBuffer[256];
void testPrint(char * inbuffer)
{
	static int y=0;
	GpTextOut(0,&testSurface, 9,y,inbuffer,0xff);
	y+=10;
}
#ifdef LOGSYSCALL
#define FUNCTION testPrint(__FUNCTION__);
#define DEBUG(x,y) sprintf(testPrintBuffer, x,y);testPrint(testPrintBuffer);
#define DEBUG2(x,y,z) sprintf(testPrintBuffer,x,y,z);testPrint(testPrintBuffer);
#else
#define FUNCTION
#define DEBUG(x,y)
#define DEBUG2(x,y,z)
#endif

struct GPFILE
{
	F_HANDLE handle;
	char * name;
	short fd;
	int inMem;

	char * writeBuffer; // Direct to SMC is very slow, use a buffer for small files
	char * writeBufferPtr;
	char * writeBufferEnd;
	char * writeBufferMemEnd;
};
#define MAXFILE 256
#define MIN_FH_NUMBER 3 
#define WRITEBUFFERINITIAL (16384) 
//start at 16k
// 0,1,2 are taken
static struct GPFILE files[MAXFILE];
static int filesInitialised = 0;

static struct GPFILE * findInMemGPFILEByName(const char * name, int setFD)
{
	int i;
	FUNCTION
	for (i=0; i<MAXFILE; ++i)
	{
		if (
			(files[i].inMem == 1) &&
			(files[i].name != 0) && 
			(strcmp(files[i].name, name) == 0)
		)
		{
			struct GPFILE * file = &files[i];
			if (setFD)
				file->fd = i+MIN_FH_NUMBER;
			return file;
		}
	}

	return 0;
}

static struct GPFILE * createGPFILE(const char * name)
{
	int i;
	int inMem = 0;
	FUNCTION
	if (!filesInitialised)
	{
		filesInitialised = 1;
		memset(files, 0, sizeof(struct GPFILE) * MAXFILE);
	}

	// In memory file? Search for one already existing
	if (strncmp(name,"mem:",4) == 0) 
	{
		struct GPFILE * file = findInMemGPFILEByName(name, 1);
		if (file)
		{
			//DEBUG("Found:%x\n", file);
			return file;
		}

		inMem = 1;
	}

	// Need a new fd, search for first empty one
	for (i=0; i<MAXFILE; ++i)
	{
		if (files[i].name == 0)
			break;
	}
	if (i==MAXFILE)
		return 0;

	struct GPFILE * file = &files[i];
	file->fd = i+MIN_FH_NUMBER;
	file->handle = -1;
	file->name = strdup(name);
	file->inMem = inMem;
	return file;
}
static void removeGPFILE(struct GPFILE * file)
{
	FUNCTION
	free(file->name);
	file->name = 0;
	if (file->writeBuffer)
	{
		free(file->writeBuffer);
		file->writeBuffer = 0;
	}
	file->fd = 0;
}

static struct GPFILE * lookupGPFILE(int fileHandle)
{
	FUNCTION
	int i = fileHandle-MIN_FH_NUMBER;
	if (i>=0 && i<MAXFILE)
		return &files[i];
	else
		return 0;
}

static int doesExist ( struct GPFILE * file )
{
	FUNCTION
	if (file->inMem)
	{
		//Aprint("Exists: %s %x\n", file->name, file->writeBuffer);
		return (file->writeBuffer!=0);
	}
	else
	{
		GPFILEATTR attr;
    
		if (GpFileAttr(file->name, &attr)!=SM_OK)
			return 0;
	}
    
	return 1;
}

static int expandWriteBuffer(struct GPFILE * file)
{
	FUNCTION
	// Increase memory
	char * newWriteBuffer = 0;
	char * currentWriteBuffer = file->writeBuffer;
	int currentMem = file->writeBufferMemEnd - file->writeBuffer;
	//Aprint("Increasing write buffer to %d\n", currentMem*2);

	// Copy old data
	newWriteBuffer = (char *) malloc(currentMem*2);
	if (!newWriteBuffer)
	{
		//Aprint("Unable to write, out of memory");
		return 0;
	}
	memcpy(newWriteBuffer, file->writeBuffer, currentMem);

	// Replace buffer
	file->writeBuffer = newWriteBuffer;
	file->writeBufferMemEnd = file->writeBuffer + currentMem*2;
	file->writeBufferPtr = file->writeBufferPtr - currentWriteBuffer + file->writeBuffer;
	file->writeBufferEnd = file->writeBufferEnd - currentWriteBuffer + file->writeBuffer;

	// Free old buffer
	free(currentWriteBuffer);

	return 1;
}

static void flushWriteBuffer(struct GPFILE * file)
{
	FUNCTION
	// Very simple buffering...
	// Once flushed it is no longer buffered here
	if (file && file->writeBuffer && !file->inMem)
	{
		//Aprint("Flushing write buffer for %s\n", file->name);

		char * writeBuffer = file->writeBuffer;
		int len = file->writeBufferEnd - file->writeBuffer;
		file->writeBuffer = 0;
		if (len)
			_write(file->fd, writeBuffer, len);

		free(writeBuffer);
	}
}

void _exit(int val)
{
	FUNCTION
	GpAppExit();
}

int _close(int fileHandle){
	FUNCTION
	struct GPFILE * file = lookupGPFILE(fileHandle);
	if (!file)
		return -1;

	flushWriteBuffer(file);

	//Aprint("Close:%s:%d\n", file->name, file->inMem);

	//DEBUG2("Close:%s:%x", file->name, file->writeBufferEnd - file->writeBuffer);
	if (!file->inMem)
	{
		ERR_CODE code = GpFileClose((F_HANDLE)file->handle);
		removeGPFILE(file);
		if (code == SM_OK)
			return 0;
		else
			return -1;
	}
	else
	{
		file->writeBufferPtr = file->writeBuffer;
	}

	return 0;
};

char * __env[1] = { 0 };
char ** environ = __env;

int _execve(char *name, char **argv, char **env)
{
	FUNCTION
	errno=ENOMEM;
	return -1;
}

int _fork()
{
	FUNCTION
	errno=EAGAIN;
	return -1;
}

int _getpid()
{
	FUNCTION
	return 1;
}

int isatty(int file)
{
	FUNCTION
	return 1;
}

int _kill(int pid, int sig){
	FUNCTION
	errno=EINVAL;
	return -1;
}

int _link(char *old, char *new){
	FUNCTION
	//FIXME? Is this symlinks? Or rename? Also what about inmem files?
	if (GpFileRename(old, new) == SM_OK)
		return 0;
	else
		return -1;
}

int _lseek(int fileHandle, int ptr, int dir)
{
	FUNCTION
	struct GPFILE * file = lookupGPFILE(fileHandle);
	ERR_CODE err;
	long oldOffset, currentOffset;
	long mode;
	if (!file)
		return -1;

	//Aprint("_lseek called for %s\n", file->name);

	if (file->writeBuffer) // Works for inmem files too...
	{
		//Aprint("Buffered seek\n");
		currentOffset = 0;
		switch (dir)
		{
		case SEEK_SET:
			currentOffset = ptr;
			break;
		case SEEK_CUR:
			currentOffset = file->writeBufferPtr - file->writeBuffer + ptr;
			break;
		case SEEK_END:
			currentOffset = file->writeBufferEnd - file->writeBuffer - ptr;
			break;
		default:
			return -1;
		}
		int len = file->writeBufferEnd - file->writeBuffer;
		if (currentOffset > len)
		{
			if ((file->writeBuffer + currentOffset) < file->writeBufferMemEnd)
				file->writeBufferEnd = file->writeBuffer+currentOffset;
			else
			{
				if (expandWriteBuffer(file))
					return _lseek(fileHandle, ptr, dir);
				else
					return -1;
			}
		}
		if (currentOffset < 0)
			currentOffset = 0;

		file->writeBufferPtr = file->writeBuffer + currentOffset;
	}
	else
	{
		//Aprint("Unbuffered seek\n");
		switch (dir)
		{
		case SEEK_SET:
			mode = FROM_BEGIN;
			break;
		case SEEK_CUR:
			mode = FROM_CURRENT;
			break;
		case SEEK_END:
			mode = FROM_END;
			ptr = -ptr;
			break;
		default:
			return -1;
		}

		err = GpFileSeek(file->handle, mode, ptr, &oldOffset);
		if (err!=SM_OK)
			return -1;

		err = GpFileSeek(file->handle, FROM_CURRENT, 0, &currentOffset);
		if (err!=SM_OK)
			return -1;
	}

	return currentOffset;	
}

int _open(const char *name, int flags, int mode)
{
	FUNCTION
	struct GPFILE * file = createGPFILE(name);
	int exists=0;

	//Aprint("_open called for %s\n", name);
	DEBUG2("OPEN:%s:%x", name, file);

	exists = doesExist(file);

	DEBUG("Exists:%x\n", exists);

	if (exists && (flags&O_TRUNC))
	{
		//DEBUG("TRUNC%d\n", 0);
		if (file->inMem)
		{
			if (file->writeBuffer)
				file->writeBufferEnd = file->writeBufferPtr = file->writeBuffer;
		}
		else
		{
			if (GpFileRemove(name) != SM_OK)
			{
				removeGPFILE(file);
				return -1;
			}

			flags |= O_CREAT;
			exists = 0;
		}
	}
	if (!exists && (flags&O_CREAT))
	{
		if (file->inMem)
		{
			file->writeBuffer = (char *)malloc(WRITEBUFFERINITIAL);
			file->writeBufferEnd = file->writeBufferPtr = file->writeBuffer;
			file->writeBufferMemEnd = file->writeBuffer + WRITEBUFFERINITIAL;
		}
		else
		{
			F_HANDLE emptyFile = 0;
			if (GpFileCreate(name, ALWAYS_CREATE, &emptyFile)!=SM_OK)
			{
				removeGPFILE(file);
				return -1;
			}

			if (GpFileClose(emptyFile)!=SM_OK)
			{
				removeGPFILE(file);
				return -1;
			}
		}
	}
	if (!file->inMem)
	{
		unsigned long readwrite = 0;
		unsigned int flagsAcc = flags&O_ACCMODE;
		switch (flagsAcc)
		{
		case O_RDWR:
			readwrite = OPEN_R|OPEN_W; 
			break;
		case O_RDONLY:
			readwrite = OPEN_R;
			break;
		case O_WRONLY:
			readwrite = OPEN_W;
			break;
		default:
			return -1;
		}

		//Aprint("FlagsAcc:%x readwrite:%x file:%s\n", flagsAcc, readwrite, file->name);

		if (GpFileOpen(name, readwrite, &file->handle)!=SM_OK)
		{
			removeGPFILE(file);
			return -1;
		}

		if (readwrite == OPEN_W) 
		{
			file->writeBuffer = (char *)malloc(WRITEBUFFERINITIAL);
			file->writeBufferEnd = file->writeBufferPtr = file->writeBuffer;
			file->writeBufferMemEnd = file->writeBuffer + WRITEBUFFERINITIAL;
			//Aprint("Creating write buffer for %s %x %x %x %x\n", file->name, flags, readwrite, OPEN_W, O_WRONLY);
		}
	}
	if (flags&O_APPEND)
	{
		flushWriteBuffer(file);
		_lseek(file->handle, 0, SEEK_END);
	}

	return file->fd;
}

int _read(int fileHandle, char * ptr, int len)
{
	FUNCTION
	ERR_CODE err;
	unsigned long read;

	struct GPFILE * file = lookupGPFILE(fileHandle);
	if (!file)
		return 0;

	//Aprint("_read called for %s\n", file->name);

	flushWriteBuffer(file);

	if (file->inMem)
	{
		//DEBUG2("Read:%s:%x\n", file->name, file->writeBufferEnd - file->writeBuffer);
		int remainingLength = (file->writeBufferEnd - file->writeBufferPtr);
		len = (len<=remainingLength) ? len : remainingLength;
		memcpy(ptr, file->writeBufferPtr, len);
		file->writeBufferPtr += len;
		//DEBUG("READ2:%x", len);
		return len;
	}
	else
	{
		err = GpFileRead(file->handle, (void *)ptr, len, &read);
		return read;
	}
}

caddr_t _sbrk(int incr){
	FUNCTION
	static int heapstart = 0;
	static int size = 0;
	int addr = size;
	if (heapstart == 0)
		heapstart = (int) gm_malloc(1024*512*10);
	
	DEBUG2("HEAPSTART:%x:%x", heapstart, addr);

	size+=incr;
	return (caddr_t)(heapstart + addr);
}

int _stat(const char *name, struct stat *buf)
{
	FUNCTION
	if (strncmp(name, "mem:", 4) == 0)
	{
		struct GPFILE * file = findInMemGPFILEByName(name, 0);
		if (!file)
			return -1;

		return _fstat(file, buf);
	}
	else
	{
		GPFILEATTR atrib;
		ERR_CODE code = GpFileAttr(name, &atrib);
		if (code != SM_OK)
			return -1;

		//Aprint("_stat called for %s\n", name);

		buf->st_size = atrib.size;
		buf->st_ino = atrib.cluster;
		//TODO:Time
		buf->st_mode = 0;
		if (atrib.attr&0x10)
			buf->st_mode |= S_IFDIR;
		if (atrib.attr&0x1)
			buf->st_mode |= S_IREAD|S_IEXEC;
		else	
			buf->st_mode |= S_IREAD|S_IWRITE|S_IEXEC;

		return 0;
	}
}

int _fstat(int fileHandle, struct stat *buf)
{
	FUNCTION
	struct GPFILE * file = lookupGPFILE(fileHandle);
	if (!file)
		return -1;

	flushWriteBuffer(file);

	if (file->inMem)
	{
		buf->st_size = file->writeBufferPtr - file->writeBuffer;
		buf->st_ino = file->writeBuffer;
		buf->st_mode = S_IREAD|S_IWRITE|S_IEXEC;
		return 0;
	}
	else
		return _stat(file->name, buf);
}

int _times(struct tms *buf){
	FUNCTION
	return -1;
}

int _unlink(char *name){
	FUNCTION
	//Aprint("REMOVE %s\n", name);
	if (strncmp(name, "mem:", 4) == 0)
	{
		struct GPFILE * file = findInMemGPFILEByName(name, 0);
		if (!file)
			return -1;

		//Aprint("unlink in memory file: %s\n", name);
		removeGPFILE(file);	
		return 0;
	}
	else
	{
		if (GpFileRemove(name)!=SM_OK)
			return -1;
		else
			return 0;
	}
}

int _wait(int *status) {
	FUNCTION
	errno=ECHILD;
	return -1;
}

int _write(int fileHandle, char *ptr, int len){
	FUNCTION
	struct GPFILE * file = lookupGPFILE(fileHandle);
	if (!file)
	{
		return 0;
	}

	DEBUG("WRITE:%x", file);

	//Aprint("_write called for %s\n", file->name);
	
	if (file->writeBuffer)
	{
		//Aprint("Buffered write\n");
		if ((file->writeBufferPtr +len) >= file->writeBufferMemEnd)
		{
			if (expandWriteBuffer(file))
				return _write(fileHandle, ptr, len);
			else
				return 0;
		}
		memcpy(file->writeBufferPtr, ptr, len);
		file->writeBufferPtr += len;
		if (file->writeBufferPtr > file->writeBufferEnd)
			file->writeBufferEnd = file->writeBufferPtr;
	}
	else
	{
		if (GpFileWrite(file->handle, (const void *)ptr, len)!=SM_OK)
			return 0;
		//Aprint("Unbuffered write!!! Len: %d\n", len);
	}

	return len;
}

int _gettimeofday(struct timeval *tv, struct timezone *tz)
{
	FUNCTION
	return -1;
}

