#include "menu_ausbjoystick.h"

#include <stdlib.h>

extern "C"
{
#include "atari.h"
#include "menu_keymap.h"
#include "atari_gp32.h"
}

#include <devicemanager.h>

static Joystick * getStick(int pos)
{
	static std::vector<Joystick *> sticks;
	if (sticks.size()<pos)
		sticks.resize(pos*2);
	Joystick * stick = sticks[pos-1];
	if (!stick)
		stick = sticks[pos-1] = &DeviceManager::instance().acquireJoystick();

	return stick;
}

AUSBJoystickMenu::AUSBJoystickMenu(int driver, const std::deque<std::string> & tokens):
	AMenu(driver),
	d_joystick(0)
{
	if (tokens.size() == 1)
		d_joystickNumber = atoi(tokens[0].c_str());
	else
		d_joystickNumber = 1;

	d_joystick = getStick(d_joystickNumber);
}

AUSBJoystickMenu::~AUSBJoystickMenu()
{
	//DeviceManager::instance().release(d_joystick);
}

void AUSBJoystickMenu::frame(int keysPressed)
{
	Joystick::DigitalPosition pos = d_joystick->getDigitalPosition(ABS_X,ABS_Y);
	bool fire = !d_joystick->getPressed().empty();
	unsigned int val = 0;
	if (pos.aPos>0)
		val|=8;
	else if (pos.aPos<0)
		val|=4;
	if (pos.bPos>0)
		val|=2;
	else if (pos.bPos<0)
		val|=1;
	val = (val<<1) | (fire?1:0);
	
	int stickNo = d_joystickNumber-1;
	while (stickNo--)
		val<<=5;

	//double x = d_joystick->getPosition(ABS_X);
	//double y = d_joystick->getPosition(ABS_Y);
	//fprintf(stdout, "me:%x val:%x aPos:%d bPos:%d fires:%d x:%f y:%f\n", this, val, pos.aPos, pos.bPos, d_joystick->getPressed().size(), x, y);

	d_joyPressed = val;
}

void AUSBJoystickMenu::clean()
{
	d_keyPressed = AKEY_NONE;
	d_operPressed = 0;
	d_consolPressed = 0;
	d_shiftPressed = 0;
	d_joyPressed = 0;
	d_5200Pressed = 0;
}

