#include "atari_gp32.h"

#include <stdlib.h>
#include <stdio.h>

#include "jpeglib.h"
#include "jpegbuffersrc.h"

#include "splash_data.h"

extern GPDRAWSURFACE testSurface, testSurface2;

// Copy a 16-bit screen to the GP's screen, adjusting brightness in the process
static void fade(unsigned int * dest, unsigned int * src, int factor)
{
	int count = 320*240*2/4;
	int i;
	unsigned int r,g,b,r2,g2,b2;
	for (i=0; i<count; ++i)
	{
		unsigned int reg = *src++;

		b = (((reg&0x3e)*factor)>>7)&0x3e;
		g = (((reg&0x7c0)*factor)>>7)&0x7c0;
		r = (((reg&0xf800)*factor)>>7)&0xf800;

		b2 = (((reg&0x3e0000)>>7)*factor)&0x3e0000;
		g2 = (((reg&0x7c00000)>>7)*factor)&0x7c00000;
		r2 = (((reg&0xf8000000)>>7)*factor)&0xf8000000;

		*dest++ = (unsigned int)r2 | (unsigned int)g2 | (unsigned int)b2| (unsigned int)r | (unsigned int)g | (unsigned int)b;
	}
}

static char * splashmem = 0;
static int splashfade = 0;
static int splashToggle = 0;
static int splashTimer = 0;
static int splashKey = 0;

void displaySplash();

void initSplash()
{
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	JSAMPARRAY buffer, buffer2;
	int row_stride;

	GpGraphicModeSet(16, NULL);
	GpLcdSurfaceGet(&testSurface, 0);
	GpLcdSurfaceGet(&testSurface2, 1);
	GpRectFill16(NULL, &testSurface, 0, 0, 320, 240, 0);

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	//jpeg_stdio_src(&cinfo, infile);
	jpeg_buffer_src(&cinfo, splash, splash_length);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	splashmem = malloc(320*240*2);

	row_stride = cinfo.output_width * cinfo.output_components;
	buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
	buffer2 = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	unsigned int * mem = (unsigned int *)splashmem;
	mem+=120;
	while (cinfo.output_scanline < cinfo.output_height) {
		jpeg_read_scanlines(&cinfo, buffer, 1);	
		jpeg_read_scanlines(&cinfo, buffer2, 1);	

		int i=0;
		unsigned char * src = buffer[0];
		unsigned char * src2 = buffer2[0];
		unsigned char r,r2,g,g2,b,b2;
		for (i=0; i<960; i+=3)
		{
			r = (src[i]&0xf8);
			r2 = (src2[i]&0xf8);
			g = (src[i+1]&0xf8);
			g2 = (src2[i+1]&0xf8);
			b = (src[i+2]&0xf8);
			b2 = (src[i+2]&0xf8);

			unsigned int col = (unsigned int)r2<<8 | (unsigned int)g2<<3 | (unsigned int)b2>>2 | (unsigned int)r<<24 | (unsigned int)g<<19 | (unsigned int)b<<14;

			*mem = col;
			mem+=120;
		}
		mem-=(120*320)+1;
	}

	jpeg_finish_decompress(&cinfo);

	displaySplash();
	splashTimer = gpNextTimer++;
	GpTimerOptSet(splashTimer, 25, 0, displaySplash);
	GpTimerSet(splashTimer);
}

void displaySplash()
{
	GPDRAWSURFACE * surface = splashToggle ? &testSurface : &testSurface2;	

	if (GpKeyGet())
	{
		splashfade = 128;
		splashKey = 1;
	}

	fade((unsigned int *)surface->ptbuffer, (unsigned int *)splashmem, splashfade);
	splashToggle ^=1;

	splashfade+=4;
	GpSurfaceFlip(surface);

	if (splashfade>128)	
		GpTimerKill(splashTimer);
}

void cleanupSplash()
{
	if (!splashKey)
		while (!GpKeyGet());
	GpTimerKill(splashTimer);
	free(splashmem);
}



